﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using Ionic.Zip;
using System.IO;
using System.Text;
using Tao.OpenGl;
using System.Xml;
using System.Windows.Forms;

namespace Rhenin
{
    public class Texture
    {
        private Image texturePath;

        public uint[] name;

        public Texture(String path)
        {
            texturePath = Image.FromFile(path);
            init();
        }

        public Texture(Image path)
        {
            texturePath = path;
            init();
        }

        private void init()
        {
            Bitmap bmp = new Bitmap(texturePath);
            //bmp.RotateFlip(RotateFlipType.RotateNoneFlipY);

            BitmapData bmpData = bmp.LockBits(new Rectangle(new Point(0 ,0), bmp.Size), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);

            byte[] imageByteArray = new byte[bmp.Width * bmp.Height * 4];
            System.Runtime.InteropServices.Marshal.Copy(bmpData.Scan0, imageByteArray, 0, imageByteArray.Length);

            bmp.UnlockBits(bmpData);

            byte r, b;
            int n;

            for (int i = 0; i < bmp.Width * bmp.Height; ++i)
            {
                n = i * 4;
                b = imageByteArray[n];
                r = imageByteArray[n + 2];

                imageByteArray[n] = r;
                imageByteArray[n + 2] = b;
            }


            name = new uint[1];
            Gl.glGenTextures(1, name);
            Gl.glBindTexture(Gl.GL_TEXTURE_RECTANGLE_ARB, name[0]);
            Gl.glPixelStorei(Gl.GL_UNPACK_ALIGNMENT, 1);

            Gl.glTexImage2D(Gl.GL_TEXTURE_RECTANGLE_ARB, 0, 4, bmp.Width, bmp.Height, 0, Gl.GL_RGBA, Gl.GL_UNSIGNED_BYTE, imageByteArray);

            Gl.glTexParameteri(Gl.GL_TEXTURE_RECTANGLE_ARB, Gl.GL_TEXTURE_MIN_FILTER, Gl.GL_NEAREST);
            Gl.glTexParameteri(Gl.GL_TEXTURE_RECTANGLE_ARB, Gl.GL_TEXTURE_MAG_FILTER, Gl.GL_NEAREST);
        }

        public void Disponse()
        {
            Gl.glDeleteTextures(1, name);
        }
    }

    public class GLUtil
    {
        /// <summary>
        /// クォータニオンの積 r <- p x q
        /// </summary>
        /// <param name="r"></param>
        /// <param name="p"></param>
        /// <param name="q"></param>
        public static void qmul(out double[] r, double[] p, double[] q)
        {
            r = new double[4];
            r[0] = p[0] * q[0] - p[1] * q[1] - p[2] * q[2] - p[3] * q[3];
            r[1] = p[0] * q[1] + p[1] * q[0] + p[2] * q[3] - p[3] * q[2];
            r[2] = p[0] * q[2] - p[1] * q[3] + p[2] * q[0] + p[3] * q[1];
            r[3] = p[0] * q[3] + p[1] * q[2] - p[2] * q[1] + p[3] * q[0];
        }
 
        /// <summary>
        /// 回転の変換行列 r <- クォータニオン q
        /// </summary>
        /// <param name="r"></param>
        /// <param name="q"></param>
        public static void qrot(out double[] r, double[] q)
        {
            r = new double[16];
            double x2 = q[1] * q[1] * 2.0;
            double y2 = q[2] * q[2] * 2.0;
            double z2 = q[3] * q[3] * 2.0;
            double xy = q[1] * q[2] * 2.0;
            double yz = q[2] * q[3] * 2.0;
            double zx = q[3] * q[1] * 2.0;
            double xw = q[1] * q[0] * 2.0;
            double yw = q[2] * q[0] * 2.0;
            double zw = q[3] * q[0] * 2.0;
  
            r[ 0] = 1.0 - y2 - z2;
            r[ 1] = xy + zw;
            r[ 2] = zx - yw;
            r[ 4] = xy - zw;
            r[ 5] = 1.0 - z2 - x2;
            r[ 6] = yz + xw;
            r[ 8] = zx + yw;
            r[ 9] = yz - xw;
            r[10] = 1.0 - x2 - y2;
            r[ 3] = r[ 7] = r[11] = r[12] = r[13] = r[14] = 0.0;
            r[15] = 1.0;
        }
    }

    public class Quaternion
    {
        public double w;
        public double x;
        public double y;
        public double z;
    }

    public class ModelLoad
    {
        public static List<GlBox> Load(String path, out Image img)
        {
            try
            {
                Image ModelTexture = null;
                String _xml = null;

                if (!File.Exists(path))
                {
                    img = null;
                    return null;
                }

                ZipFile zip = ZipFile.Read(path);
                foreach (ZipEntry entry in zip)
                {
                    var stream = entry.OpenReader();
                    byte[] buffer = new byte[stream.Length];
                    stream.Read(buffer, 0, buffer.Length);
                    stream.Close();
                    if (entry.FileName.Contains(".png"))
                    {
                        ImageConverter ic = new ImageConverter();
                        ModelTexture = (Image)ic.ConvertFrom(buffer);
                    }
                    else if (entry.FileName.Contains(".xml"))
                    {
                        _xml = Encoding.UTF8.GetString(buffer);
                    }
                }
                zip.Dispose();
                List<GlBox> mm = XmlLoad(_xml);
                img = ModelTexture;
                return mm;
            }
            catch (Exception e)
            {

            }
            img = null;
            return null;
        }

        private static List<GlBox> XmlLoad(String file)
        {
            XmlDocument doc = new XmlDocument();
            doc.Load(new StringReader(file));
            List<GlBox> boxes = new List<GlBox>();
            int index = 0;
            foreach (XmlNode node in doc.GetElementsByTagName("Position"))
            {
                string innerText = node.InnerText;
                boxes.Add(new GlBox());
                String[] th = innerText.Split(',');
                boxes[index].x = Double.Parse(th[0]);
                boxes[index].y = Double.Parse(th[1]);
                boxes[index].z = Double.Parse(th[2]);
                boxes[index].name = "box" + (index + 1);
                index++;
            }
            index = 0;
            foreach (XmlNode node in doc.GetElementsByTagName("Offset"))
            {
                string innerText = node.InnerText;
                String[] th = innerText.Split(',');
                boxes[index].offsetX = Double.Parse(th[0]);
                boxes[index].offsetY = Double.Parse(th[1]);
                boxes[index].offsetZ = Double.Parse(th[2]);
                index++;
            }
            index = 0;
            foreach (XmlNode node in doc.GetElementsByTagName("Size"))
            {
                string innerText = node.InnerText;
                String[] th = innerText.Split(',');
                boxes[index].width = Int32.Parse(th[0]);
                boxes[index].height = Int32.Parse(th[1]);
                boxes[index].depth = Int32.Parse(th[2]);
                index++;
            }
            index = 0;
            foreach (XmlNode node in doc.GetElementsByTagName("Rotation"))
            {
                string innerText = node.InnerText;
                String[] th = innerText.Split(',');
                boxes[index].angleX = Double.Parse(th[0]);
                boxes[index].angleY = Double.Parse(th[1]);
                boxes[index].angleZ = Double.Parse(th[2]);
                index++;
            }
            index = 0;
            foreach (XmlNode node in doc.GetElementsByTagName("TextureOffset"))
            {
                string innerText = node.InnerText;
                String[] th = innerText.Split(',');
                boxes[index].texturePoint.X = Int32.Parse(th[0]);
                boxes[index].texturePoint.Y = Int32.Parse(th[1]);
                index++;
            }

            return boxes;
        }

        public static List<GlBox> LoadRHN(String path, out Image img, out String[] pars)
        {
            pars = null;
            img = null;
            Image ModelTexture = null;
            BinaryReader br = new BinaryReader(new FileStream(path, FileMode.Open));
            String head = br.ReadString();
            if (head.Equals("RHENIN"))
            {
                int boxCount = br.ReadInt32();
                List<GlBox> boxs = new List<GlBox>();
                int textureW = br.ReadInt32();
                int textureH = br.ReadInt32();
                string extend = br.ReadString();
                string package = br.ReadString();
                string modelName = br.ReadString();
                string import = br.ReadString();
                bool isImage = br.ReadBoolean();
                for (int i = 0; i < boxCount; i++)
                {
                    GlBox box = new GlBox();
                    boxs.Add(box);
                    boxs[i].name = br.ReadString();
                    boxs[i].width = br.ReadDouble();
                    boxs[i].height = br.ReadDouble();
                    boxs[i].depth = br.ReadDouble();
                    boxs[i].x = br.ReadDouble();
                    boxs[i].y = br.ReadDouble();
                    boxs[i].z = br.ReadDouble();
                    boxs[i].offsetX = br.ReadDouble();
                    boxs[i].offsetY = br.ReadDouble();
                    boxs[i].offsetZ = br.ReadDouble();
                    boxs[i].angleX = br.ReadDouble();
                    boxs[i].angleY = br.ReadDouble();
                    boxs[i].angleZ = br.ReadDouble();
                    boxs[i].texturePoint.X = br.ReadInt32();
                    boxs[i].texturePoint.Y = br.ReadInt32();
                    boxs[i].scale = br.ReadDouble();
                    boxs[i].mirror = br.ReadBoolean();
                }
                String oreshura = br.ReadString();
                pars = new String[] { package, modelName, import, extend };
                if (oreshura.Equals("俺の彼女と幼なじみが修羅場すぎる"))
                {
                    if (isImage)
                    {
                        int bufferSize = br.ReadInt32();
                        int seek = (int)br.ReadInt64();
                        byte[] buf = new byte[bufferSize];
                        for (int i = 0; i < bufferSize; i++)
                        {
                            byte[] b = new byte[(seek + 1) + i];
                            int q = br.Read(b, seek + i, 1);
                            buf[i] = b[b.Length - 1];
                        }
                        ModelTexture = ByteArrayToImage(buf);
                        img = ModelTexture;
                        return boxs;
                    }
                    return boxs;
                }
            }
            br.Close();
            return null;
        }

        public static Image ByteArrayToImage(byte[] b)
        {
            ImageConverter imgconv = new ImageConverter();
            Image img = (Image)imgconv.ConvertFrom(b);
            return img;
        }
    }

    public class ModelSave
    {
        static ImageConverter imgconv = new ImageConverter();

        public static void saveRhn(String path, List<GlBox> boxList, ListBox list, Size textureSize, String extend, Image img, String[] args)
        {
            BinaryWriter fs = new BinaryWriter(new FileStream(path, FileMode.Create));
            fs.Write("RHENIN");
            fs.Write(boxList.Count);
            fs.Write(textureSize.Width);
            fs.Write(textureSize.Height);
            fs.Write(extend);
            fs.Write(args[0]);  //package
            fs.Write(args[1]);  //ModelName
            fs.Write(args[2]);  //Import
            fs.Write(img != null);
            for (int i = 0; i < boxList.Count; i++)
            {
                fs.Write(list.Items[i].ToString());
                fs.Write(boxList[i].width);
                fs.Write(boxList[i].height);
                fs.Write(boxList[i].depth);
                fs.Write(boxList[i].x);
                fs.Write(boxList[i].y);
                fs.Write(boxList[i].z);
                fs.Write(boxList[i].offsetX);
                fs.Write(boxList[i].offsetY);
                fs.Write(boxList[i].offsetZ);
                fs.Write(boxList[i].angleX);
                fs.Write(boxList[i].angleY);
                fs.Write(boxList[i].angleZ);
                fs.Write(boxList[i].texturePoint.X);
                fs.Write(boxList[i].texturePoint.Y);
                fs.Write(boxList[i].scale);
                fs.Write(boxList[i].mirror);
            }
            fs.Write("俺の彼女と幼なじみが修羅場すぎる");
            if (img != null)
            {
                byte[] imageBuf = ImageToByteArray(img);
                fs.Write(imageBuf.Length);
                long currentSeek = fs.Seek(0, SeekOrigin.Current);
                fs.Write(currentSeek + 8);
                fs.Write(imageBuf);
            }
            fs.Close();
        }

        public static byte[] ImageToByteArray(Image img)
        {
            byte[] b = (byte[])imgconv.ConvertTo(img, typeof(byte[]));
            return b;
        }
    }

    public class CustomModel
    {
        private List<GlBox> boxs;
        public Image image;

        public CustomModel(List<GlBox> par1)
        {
            boxs = par1;
        }

        public void Render()
        {
            for (int i = 0; i < boxs.Count; i++)
            {
                boxs[i].render();
            }
        }
    }

    public abstract class Model
    {
        abstract public void render();
    }

    public class ModelCustom : Model
    {
        public List<GlBox> boxs;
        public String[] nameTable;
        public Hashtable rotateMode;
        public Hashtable rotateAngle;
        public Hashtable defaultAngle;
        public Hashtable optionPane;

        public int motionMode;
        private int frame;

        public override void render()
        {
            setLoationAngle();
            for (int i = 0; i < boxs.Count; i++)
            {
                if (optionPane.Contains(nameTable[i]))
                {
                    if (optionPane[nameTable[i]].Equals("b"))
                    {
                        Gl.glDepthMask(Gl.GL_FALSE);
                        boxs[i].render();
                        Gl.glDepthMask(Gl.GL_TRUE);
                    }
                    else
                    {
                        boxs[i].render();
                    }
                }
                else
                {
                    boxs[i].render();
                }
            }
        }

        public void result()
        {
            for (int i = 0; i < boxs.Count; i++)
            {
                if (defaultAngle.Contains(nameTable[i]))
                {
                    Vector3 vec = (Vector3)defaultAngle[nameTable[i]];
                    boxs[i].setAngle(vec.angleX, vec.angleY, vec.angleZ);
                }
                else
                {
                    boxs[i].setAngle(0, 0, 0);
                }
            }
        }

        private void setLoationAngle()
        {
            if (motionMode == 1)
            {

                frame += 10;
                for (int i = 0; i < boxs.Count; i++)
                {
                    if (rotateMode.Contains(nameTable[i]))
                    {
                        double time = Math.Sin(frame * Math.PI / 180) * (int)rotateAngle[nameTable[i]];
                        int mode = (int)rotateMode[nameTable[i]];
                        if (mode == 1)
                            boxs[i].setAngle(time, 0, 0);
                        if (mode == 2)
                            boxs[i].setAngle(-time, 0, 0);
                        if (mode == 3)
                            boxs[i].setAngle(0, time, 0);
                        if (mode == 4)
                            boxs[i].setAngle(0, -time, 0);
                    }
                }
            }
            if (motionMode == 2)
            {
                frame += 10;
                for (int i = 0; i < boxs.Count; i++)
                {
                    boxs[i].setAngle(frame % 360 - 180, frame % 360 - 180, frame % 360 - 180);
                }
            }
        }
    }

    public class Vector3
    {
        public double angleX;
        public double angleY;
        public double angleZ;
        public Vector3(double par1, double par2, double par3)
        {
            angleX = par1;
            angleY = par2;
            angleZ = par3;
        }
    }

    public class ModelBiped : Model
    {
        public GlBox head;
        public GlBox headwear;
        public GlBox body;
        public GlBox rightArm;
        public GlBox leftArm;
        public GlBox rightLeg;
        public GlBox leftLeg;

        public int motionMode;
        private int frame;

        public ModelBiped()
        {
            head = new GlBox(8.0, 8.0, 8.0, 0.0, 0.0, 0.0, -4.0, -8.0, -4.0);
            head.texturePoint = new Point(0, 0);

            headwear = new GlBox(8.0, 8.0, 8.0, 0.0, 0.0, 0.0, -4.0, -8.0, -4.0, 1.1);
            headwear.texturePoint = new Point(32, 0);

            body = new GlBox(8.0, 12.0, 4.0, 0.0, 0.0, 0.0, -4.0, 0.0, -2.0);
            body.texturePoint = new Point(16, 16);

            rightArm = new GlBox(4.0, 12.0, 4.0, -5.0, 2.0, 0.0, -3.0, -2.0, -2.0);
            rightArm.texturePoint = new Point(40, 16);

            leftArm = new GlBox(4.0, 12.0, 4.0, 5.0, 2.0, 0.0, -1.0, -2.0, -2.0);
            leftArm.texturePoint = new Point(40, 16);
            leftArm.mirror = true;

            rightLeg = new GlBox(4.0, 12.0, 4.0, -2.0, 12.0, 0.0, -2.0, 0.0, -2.0);
            rightLeg.texturePoint = new Point(0, 16);

            leftLeg = new GlBox(4.0, 12.0, 4.0, 2.0, 12.0, 0.0, -2.0, 0.0, -2.0);
            leftLeg.texturePoint = new Point(0, 16);
            leftLeg.mirror = true;

        }

        public override void render()
        {
            setLoationAngle();
            rightLeg.render();
            leftLeg.render();
            rightArm.render();
            body.render();
            leftArm.render();
            head.render();
            Gl.glDepthMask(Gl.GL_FALSE);
            headwear.render();
            Gl.glDepthMask(Gl.GL_TRUE);
        }

        private void setLoationAngle()
        {
            if (motionMode == 1)
            {
                frame += 10;
                double time = Math.Sin(frame * Math.PI / 180) * 50;
                leftArm.setAngle(time, 0, 0);
                rightArm.setAngle(-time, 0, 0);
                leftLeg.setAngle(-time, 0, 0);
                rightLeg.setAngle(time, 0, 0);
            }
            if (motionMode == 2)
            {
                frame += 10;
                head.setAngle(frame % 360, frame % 360, frame % 360);
                headwear.setAngle(frame % 360, frame % 360, frame % 360);
                body.setAngle(frame % 360, frame % 360, frame % 360);
                rightArm.setAngle(frame % 360, frame % 360, frame % 360);
                leftArm.setAngle(frame % 360, frame % 360, frame % 360);
                rightLeg.setAngle(frame % 360, frame % 360, frame % 360);
                leftLeg.setAngle(frame % 360, frame % 360, frame % 360);
            }
        }
    }

    public class GlBox
    {
        public double width;
        public double height;
        public double depth;

        public double x;
        public double y;
        public double z;

        public double offsetX;
        public double offsetY;
        public double offsetZ;

        public double angleX;
        public double angleY;
        public double angleZ;

        public double scale;

        public Point texturePoint = new Point(0, 0);

        public bool mirror = false;

        public string name;

        //0=通常,1=線,2非表示
        public int drawMode = 0;

        /// <summary>
        /// 新しい箱のインスタンスを生成します
        /// </summary>
        public GlBox()
        {

        }
        /// <summary>
        /// 新しい箱のインスタンスを生成します
        /// </summary>
        /// <param name="par1">横のサイズ</param>
        /// <param name="par2">縦のサイズ</param>
        /// <param name="par3">奥行きのサイズ</param>
        public GlBox(double par1, double par2, double par3)
        {
            width = par1;
            height = par2;
            depth = par3;
            x = 0;
            y = 0;
            z = 0;
        }

        /// <summary>
        /// 新しい箱のインスタンスを生成します
        /// </summary>
        /// <param name="par1">横のサイズ</param>
        /// <param name="par2">縦のサイズ</param>
        /// <param name="par3">奥行きのサイズ</param>
        /// <param name="par4">X座標</param>
        /// <param name="par5">Y座標</param>
        /// <param name="par6">Z座標</param>
        /// <param name="par7">Xオフセット</param>
        /// <param name="par8">Yオフセット</param>
        /// <param name="par9">Zオフセット</param>
        public GlBox(double par1, double par2, double par3, double par4, double par5, double par6, double par7, double par8, double par9)
        {
            width = par1;
            height = par2;
            depth = par3;
            x = par4;
            y = par5;
            z = par6;
            offsetX = par7;
            offsetY = par8;
            offsetZ = par9;
        }

        /// <summary>
        /// 新しい箱のインスタンスを生成します
        /// </summary>
        /// <param name="par1">横のサイズ</param>
        /// <param name="par2">縦のサイズ</param>
        /// <param name="par3">奥行きのサイズ</param>
        /// <param name="par4">X座標</param>
        /// <param name="par5">Y座標</param>
        /// <param name="par6">Z座標</param>
        /// <param name="par7">Xオフセット</param>
        /// <param name="par8">Yオフセット</param>
        /// <param name="par9">Zオフセット</param>
        /// <param name="par10">拡大率</param>
        public GlBox(double par1, double par2, double par3, double par4, double par5, double par6, double par7, double par8, double par9, double par10)
        {
            width = par1 + par10;
            height = par2 + par10;
            depth = par3 + par10;
            x = par4;
            y = par5;
            z = par6;
            offsetX = par7 - par10;
            offsetY = par8 - par10;
            offsetZ = par9 - par10;
            scale = par10;
        }

        public void setAngle(double _x, double _y, double _z)
        {
            angleX = _x;
            angleY = _y;
            angleZ = _z;
        }

        public void render(bool par = true)
        {

            if (drawMode == 2)
                return;

            if (drawMode == 0)
            {
                Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);
            }
            else if(drawMode == 1)
            {
                Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_LINE);
            }
              double[][] vertex = new Double[][]{
              new Double[]{ 0.0,  0.0,  0.0 }, /* A */
              new Double[]{ width + scale, 0.0,  0.0 }, /* B */
              new Double[]{ width + scale, height + scale, 0.0 }, /* C */
              new Double[]{ 0.0,  height + scale, 0.0 }, /* D */
              new Double[]{ 0.0,  0.0,  depth + scale}, /* E */
              new Double[]{ width + scale, 0.0,  depth + scale }, /* F */
              new Double[]{ width + scale, height + scale, depth + scale }, /* G */
              new Double[]{ 0.0,  height + scale, depth + scale }  /* H */
            };

            if (par)
            {
                Gl.glPushMatrix();
                Form1.GlInitPush();
            }

            Gl.glTranslated((x + (offsetX - scale)), (y + (offsetY - scale)), (z + (offsetZ - scale)));
            Gl.glTranslated(-offsetX - scale, -offsetY - scale, -offsetZ - scale);
            Gl.glRotated(angleX, 1.0, 0.0, 0.0);
            Gl.glRotated(angleY, 0.0, 1.0, 0.0);
            Gl.glRotated(angleZ, 0.0, 0.0, 1.0);

            Gl.glTranslated(offsetX - scale, offsetY - scale, offsetZ - scale);

            
            Gl.glBlendFunc(Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA);
            Gl.glEnable(Gl.GL_TEXTURE_RECTANGLE_ARB);
            Gl.glBegin(Gl.GL_QUADS);
            for (int j = 0; j < 6; ++j)
            {
                for (int i = 0; i < 4; ++i)
                {
                    getTexCoord(i, j);
                    //Gl.glTexCoord2d(j * 10.0, i * 20.0);
                    //Gl.glColor3dv(color[j]);
                    //if(Form1.mainOption.ShowShadow)
                    //    Gl.glNormal3dv(normal[j]);
                    Gl.glVertex3dv(vertex[face[j][i]]);
                }
            }
            Gl.glEnd();
            Gl.glDisable(Gl.GL_TEXTURE_RECTANGLE_ARB);
            Gl.glEnable(Gl.GL_BLEND);
            if (par)
            {
                Gl.glPopMatrix();
                Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);
            }
        }

        private void getTexCoord(int i, int j)
        {
            if (!mirror)
            {
                if (j == 0)
                {
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + depth);
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + height + depth);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + height + depth);
                }
                else if (j == 1)
                {
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + depth);
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width, texturePoint.Y + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width, texturePoint.Y + height + depth);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + height + depth);
                }
                else if (j == 2)
                {
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width, texturePoint.Y + depth);
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width * 2, texturePoint.Y + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width * 2, texturePoint.Y + height + depth);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width, texturePoint.Y + height + depth);
                }
                else if (j == 3)
                {
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + 0, texturePoint.Y + depth);
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + height + depth);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + 0, texturePoint.Y + height + depth);
                }
                else if (j == 4)
                {
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y);
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + depth);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + depth);
                }
                else if (j == 5)
                {
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth + width * 2, texturePoint.Y);
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth + width * 2, texturePoint.Y + depth);
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + depth);
                }
            }
            else if (mirror)
            {
                if (j == 0)
                {
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + depth);
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + depth);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + height + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + height + depth);
                }
                else if (j == 3)
                {
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + depth);
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width, texturePoint.Y + depth);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width, texturePoint.Y + height + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + height + depth);
                }
                else if (j == 2)
                {
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width, texturePoint.Y + depth);
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width * 2, texturePoint.Y + depth);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width * 2, texturePoint.Y + height + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth * 2 + width, texturePoint.Y + height + depth);
                }
                else if (j == 1)
                {
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + 0, texturePoint.Y + depth);
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + depth);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + height + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + 0, texturePoint.Y + height + depth);
                }
                else if (j == 4)
                {
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y);
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth, texturePoint.Y + depth);
                }
                else if (j == 5)
                {
                    if (i == 1)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y);
                    if (i == 0)
                        Gl.glTexCoord2d(texturePoint.X + depth + width * 2, texturePoint.Y);
                    if (i == 3)
                        Gl.glTexCoord2d(texturePoint.X + depth + width * 2, texturePoint.Y + depth);
                    if (i == 2)
                        Gl.glTexCoord2d(texturePoint.X + depth + width, texturePoint.Y + depth);
                }
            }
        }

        public Image getImage()
        {
            int _width = (int)depth * 2 + (int)width * 2;
            int _height = (int)depth + (int)height;
            if (_width == 0 || _height == 0)
            {
                return null;
            }
            Bitmap bmp = new Bitmap(_width, _height);
            Graphics g = Graphics.FromImage(bmp);
            g.FillRectangle(Brushes.Red, new Rectangle((int)depth, 0, (int)width, (int)depth));
            g.FillRectangle(Brushes.DarkRed, new Rectangle((int)depth + (int)width, 0, (int)width, (int)depth));
            g.FillRectangle(Brushes.Green, new Rectangle(0, (int)depth, (int)depth, (int)height));
            g.FillRectangle(Brushes.DarkGreen, new Rectangle((int)depth, (int)depth, (int)width, (int)height));
            g.FillRectangle(Brushes.Blue, new Rectangle((int)depth + (int)width, (int)depth, (int)depth, (int)height));
            g.FillRectangle(Brushes.DarkBlue, new Rectangle((int)depth * 2 + (int)width, (int)depth, (int)width, (int)height));
            return bmp;
        }
        int[][] edge = new Int32[][]{
                new int[]{ 0, 1 }, /* ア (A-B) */
                new int[]{ 1, 2 }, /* イ (B-C) */
                new int[]{ 2, 3 }, /* ウ (C-D) */
                new int[]{ 3, 0 }, /* エ (D-A) */
                new int[]{ 4, 5 }, /* オ (E-F) */
                new int[]{ 5, 6 }, /* カ (F-G) */
                new int[]{ 6, 7 }, /* キ (G-H) */
                new int[]{ 7, 4 }, /* ク (H-E) */
                new int[]{ 0, 4 }, /* ケ (A-E) */
                new int[]{ 1, 5 }, /* コ (B-F) */
                new int[]{ 2, 6 }, /* サ (C-G) */
                new int[]{ 3, 7 }  /* シ (D-H) */
            };

        int[][] face = new Int32[][]{
            new int[]{ 0, 1, 2, 3 }, /* A-B-C-D を結ぶ面 */
            new int[]{ 1, 5, 6, 2 }, /* B-F-G-C を結ぶ面 */
            new int[]{ 5, 4, 7, 6 }, /* F-E-H-G を結ぶ面 */
            new int[]{ 4, 0, 3, 7 }, /* E-A-D-H を結ぶ面 */
            new int[]{ 4, 5, 1, 0 }, /* E-F-B-A を結ぶ面 */
            new int[]{ 3, 2, 6, 7 }  /* D-C-G-H を結ぶ面 */
        };


        double[][] color = new Double[][]{
          new Double[]{ 1.0, 0.0, 0.0 }, /* 赤 */
          new Double[]{ 0.0, 1.0, 0.0 }, /* 緑 */
          new Double[]{ 0.0, 0.0, 1.0 }, /* 青 */
          new Double[]{ 1.0, 1.0, 0.0 }, /* 黄 */
          new Double[]{ 1.0, 0.0, 1.0 }, /* マゼンタ */
          new Double[]{ 0.0, 1.0, 1.0 }  /* シアン 　*/
        };

        double[][] faceTex = new Double[][]{
            new Double[]{ 0.0, 0.0},
            new Double[]{ 0.0, 0.0},
            new Double[]{ 0.0, 0.0},
            new Double[]{ 0.0, 0.0},
            new Double[]{ 0.0, 0.0},
            new Double[]{ 0.0, 0.0}
        };

        double[][] normal = new Double[][]{
          new Double[]{ 0.0, 0.0,-1.0 },
          new Double[]{ 1.0, 0.0, 0.0 },
          new Double[]{ 0.0, 0.0, 1.0 },
          new Double[]{-1.0, 0.0, 0.0 },
          new Double[]{ 0.0,-1.0, 0.0 },
          new Double[]{ 0.0, 1.0, 0.0 }
        };
    }
}
